@@ -15,7 +15,13 @@
 
 #include <postgres.h>	// Needs to be first.
 
+#if defined(__APPLE__)
+#include <libkern/OSByteOrder.h>
+#define bswap_64 OSSwapInt64
+#else
 #include <byteswap.h>
+#endif
+
 #include <funcapi.h>
 #include <math.h>
 #include <stdlib.h>
@@ -23,8 +29,15 @@
 #include <inttypes.h>
 #include "utils/array.h"
 #include "utils/bytea.h"
+#if (PG_VERSION_NUM >= 100000)
+#include "utils/fmgrprotos.h"
+#endif
+#include "utils/int8.h"
+#include "utils/lsyscache.h"
 #include "utils/memutils.h"
 #include "catalog/pg_type.h"
+#include "lib/stringinfo.h"
+#include "libpq/pqformat.h"
 
 #include "MurmurHash3.h"
 
@@ -361,31 +374,70 @@ typedef struct
 static void
 bitstream_pack(bitstream_write_cursor_t * bwcp, uint32_t val)
 {
-    // Fetch the quadword where our data goes.
-    uint64_t qw = * (uint64_t *) bwcp->bwc_curp;
+    size_t bits_left_in_first_byte = 8 - bwcp->bwc_used;
+    size_t bits_to_write_after_first_byte = bwcp->bwc_nbits - bits_left_in_first_byte;
+    size_t full_bytes_to_write = bits_to_write_after_first_byte / 8;
+    size_t remainder_bits_after_full_bytes = bits_to_write_after_first_byte % 8;
 
-    // Swap the bytes.
-    qw = bswap_64(qw);
-
-    // Shift our bits into place and combine.
-    qw |= ((uint64_t) val << (64 - bwcp->bwc_nbits - bwcp->bwc_used));
-
-    // Swap the bytes back again.
-    qw = bswap_64(qw);
-
-    // Write the word back out.
-    * (uint64_t *) bwcp->bwc_curp = qw;
-
-    // We've used some more bits now.
-    bwcp->bwc_used += bwcp->bwc_nbits;
-
-    // Normalize the cursor.
-    while (bwcp->bwc_used >= 8)
+    // write is small enough that it fits in current byte's remaining space with
+    // room to spare, so pad the bottom of the byte by left-shifting
+    if (bwcp->bwc_nbits < bits_left_in_first_byte)
     {
-        bwcp->bwc_used -= 8;
-        bwcp->bwc_curp += 1;
+        * bwcp->bwc_curp = (* bwcp->bwc_curp) | ((uint8_t)(val << (bits_left_in_first_byte - bwcp->bwc_nbits)));
+
+        // consume part of the byte and exit
+        bwcp->bwc_used += bwcp->bwc_nbits;
+        return;
     }
 
+    // write fits exactly in current byte's remaining space, so just OR it in to
+    // the bottom bits of the byte
+    if (bwcp->bwc_nbits == bits_left_in_first_byte)
+    {
+        * bwcp->bwc_curp = (* bwcp->bwc_curp) | (uint8_t)(val);
+
+        // consume remainder of byte and exit
+        bwcp->bwc_used = 0;
+        bwcp->bwc_curp += 1;
+        return;
+    }
+
+    // write DOES NOT fit into current byte, so shift off all but the topmost
+    // bits from the value and OR those into the bottom bits of the byte
+    /* bwcp->bwc_nbits > bits_left_in_first_byte */
+    * bwcp->bwc_curp = (* bwcp->bwc_curp) | ((uint8_t)(val >> (bwcp->bwc_nbits - bits_left_in_first_byte)));
+
+    // consume remainder of byte
+    bwcp->bwc_used = 0;
+    bwcp->bwc_curp += 1;
+
+    // if there are 8 or more bits of the value left to write, write them in one
+    // byte chunks, higher chunks first
+    if (full_bytes_to_write > 0)
+    {
+        for (size_t i = 0; i < full_bytes_to_write; ++i)
+        {
+            size_t bits_to_keep = bits_left_in_first_byte + (8 * (i + 1));
+            size_t right_shift = (bwcp->bwc_nbits - bits_to_keep);
+
+            // no OR here because byte is guaranteed to be completely unused
+            // see above, before conditional
+            * bwcp->bwc_curp = (uint8_t)(val >> right_shift);
+
+            // consume entire byte
+            bwcp->bwc_used = 0;
+            bwcp->bwc_curp += 1;
+        }
+    }
+    if (remainder_bits_after_full_bytes > 0)
+    {
+        uint8_t mask = (1 << remainder_bits_after_full_bytes) - 1;
+        // no OR here because byte is guaranteed to be completely unused
+        * bwcp->bwc_curp = ((uint8_t)val & mask) << (8 - remainder_bits_after_full_bytes);
+
+        // consume part of the byte
+        bwcp->bwc_used = remainder_bits_after_full_bytes;
+    }
 }
 
 static void
@@ -505,7 +557,7 @@ check_metadata(multiset_t const * i_omp, multiset_t const * i_imp)
         ereport(ERROR,
                 (errcode(ERRCODE_DATA_EXCEPTION),
                  errmsg("explicit threshold does not match: "
-                        "source uses %ld and dest uses %ld",
+                        "source uses " INT64_FORMAT " and dest uses " INT64_FORMAT,
                         i_imp->ms_expthresh, i_omp->ms_expthresh)));
     }
 
@@ -625,9 +677,9 @@ multiset_tostring(multiset_t const * i_msp)
     // with the automatically determined value.
     //
     if (expthresh == -1)
-        snprintf(expbuf, sizeof(expbuf), "%ld(%ld)", expthresh, expval);
+        snprintf(expbuf, sizeof(expbuf), INT64_FORMAT "(%ld)", expthresh, expval);
     else
-        snprintf(expbuf, sizeof(expbuf), "%ld", expthresh);
+        snprintf(expbuf, sizeof(expbuf), INT64_FORMAT, expthresh);
 
     // Allocate an initial return buffer.
     len = 1024;
@@ -1722,7 +1774,7 @@ hll_typmod_out(PG_FUNCTION_ARGS)
     char * typmodstr;
 
     memset(buffer, '\0', sizeof(buffer));
-    snprintf(buffer, sizeof(buffer), "(%d,%d,%ld,%d)",
+    snprintf(buffer, sizeof(buffer), "(%d,%d," INT64_FORMAT ",%d)",
              log2m, regwidth, expthresh, sparseon);
 
     len = strlen(buffer) + 1;
@@ -1918,7 +1970,7 @@ multiset_card(multiset_t const * i_msp)
             else if (estimator <= large_estimator_cutoff)
                 retval = estimator;
             else
-                return (-1 * two_to_l) * log(1.0 - (estimator/two_to_l));
+                return (-1 * (int64_t)two_to_l) * log(1.0 - (estimator/two_to_l));
         }
         break;
 
@@ -2219,13 +2271,12 @@ hll_type(PG_FUNCTION_ARGS)
     size_t asz;
     multiset_t	msa;
     uint8_t type;
-    uint8_t vers;
 
     ab = PG_GETARG_BYTEA_P(0);
     asz = VARSIZE(ab) - VARHDRSZ;
 
     // Unpack the multiset.
-    vers = multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, &type);
+    multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, &type);
 
 	PG_RETURN_INT32(type);
 }
@@ -2240,13 +2291,12 @@ hll_log2m(PG_FUNCTION_ARGS)
     bytea * ab;
     size_t asz;
     multiset_t	msa;
-    uint8_t vers;
 
     ab = PG_GETARG_BYTEA_P(0);
     asz = VARSIZE(ab) - VARHDRSZ;
 
     // Unpack the multiset.
-    vers = multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, NULL);
+    multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, NULL);
 
 	PG_RETURN_INT32(msa.ms_log2nregs);
 }
@@ -2261,13 +2311,12 @@ hll_regwidth(PG_FUNCTION_ARGS)
     bytea * ab;
     size_t asz;
     multiset_t	msa;
-    uint8_t vers;
 
     ab = PG_GETARG_BYTEA_P(0);
     asz = VARSIZE(ab) - VARHDRSZ;
 
     // Unpack the multiset.
-    vers = multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, NULL);
+    multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, NULL);
 
 	PG_RETURN_INT32(msa.ms_nbits);
 }
@@ -2282,7 +2331,6 @@ hll_expthresh(PG_FUNCTION_ARGS)
     bytea * ab;
     size_t asz;
     multiset_t	msa;
-    uint8_t vers;
 
     size_t nbits;
     size_t nregs;
@@ -2296,7 +2344,7 @@ hll_expthresh(PG_FUNCTION_ARGS)
     asz = VARSIZE(ab) - VARHDRSZ;
 
     // Unpack the multiset.
-    vers = multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, NULL);
+    multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, NULL);
 
     nbits = msa.ms_nbits;
     nregs = msa.ms_nregs;
@@ -2342,13 +2390,12 @@ hll_sparseon(PG_FUNCTION_ARGS)
     bytea * ab;
     size_t asz;
     multiset_t	msa;
-    uint8_t vers;
 
     ab = PG_GETARG_BYTEA_P(0);
     asz = VARSIZE(ab) - VARHDRSZ;
 
     // Unpack the multiset.
-    vers = multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, NULL);
+    multiset_unpack(&msa, (uint8_t *) VARDATA(ab), asz, NULL);
 
 	PG_RETURN_INT32(msa.ms_sparseon);
 }
@@ -2564,6 +2611,70 @@ hll_hash_varlena(PG_FUNCTION_ARGS)
     PG_RETURN_INT64(out[0]);
 }
 
+
+// Hash any scalar data type.
+//
+PG_FUNCTION_INFO_V1(hll_hash_any);
+Datum		hll_hash_any(PG_FUNCTION_ARGS);
+Datum
+hll_hash_any(PG_FUNCTION_ARGS)
+{
+    Datum keyDatum = PG_GETARG_DATUM(0);
+    Datum seedDatum = PG_GETARG_DATUM(1);
+    Datum hashResultDatum = 0;
+
+    Oid keyTypeId = get_fn_expr_argtype(fcinfo->flinfo, 0);
+    int16 keyTypeLength = get_typlen(keyTypeId);
+
+    /* dispatch to corresponding hash function for key type */
+    switch (keyTypeLength)
+    {
+    case 1:
+        hashResultDatum = DirectFunctionCall2(hll_hash_1byte, keyDatum, seedDatum);
+        break;
+
+    case 2:
+        hashResultDatum = DirectFunctionCall2(hll_hash_2byte, keyDatum, seedDatum);
+        break;
+
+    case 4:
+        hashResultDatum = DirectFunctionCall2(hll_hash_4byte, keyDatum, seedDatum);
+        break;
+
+    case 8:
+        hashResultDatum = DirectFunctionCall2(hll_hash_8byte, keyDatum, seedDatum);
+        break;
+
+    case -1:
+    case -2:
+        hashResultDatum = DirectFunctionCall2(hll_hash_varlena, keyDatum, seedDatum);
+        break;
+
+    default:
+    {
+        /*
+         * We have a fixed-size type such as char(10), macaddr, circle, etc. We
+         * first convert this type to its variable-length binary representation
+         * and then dispatch to the variable-length hashing function.
+         */
+        Oid keyTypeSendFunction = InvalidOid;
+        bool keyTypeVarlena = false;
+        Datum keyBinaryDatum = 0;
+
+        /* no need to worry about SPI for these types' output functions */
+        getTypeBinaryOutputInfo(keyTypeId, &keyTypeSendFunction, &keyTypeVarlena);
+        keyBinaryDatum = OidFunctionCall1(keyTypeSendFunction, keyDatum);
+
+        hashResultDatum = DirectFunctionCall2(hll_hash_varlena, keyBinaryDatum,
+                                              seedDatum);
+        break;
+    }
+    }
+
+    PG_RETURN_INT64(hashResultDatum);
+}
+
+
 PG_FUNCTION_INFO_V1(hll_eq);
 Datum		hll_eq(PG_FUNCTION_ARGS);
 Datum
@@ -2742,10 +2853,26 @@ hll_add_trans4(PG_FUNCTION_ARGS)
                 (errcode(ERRCODE_DATA_EXCEPTION),
                  errmsg("hll_add_trans4 outside transition context")));
 
-    // Is the first argument a NULL?
+    // If the first argument is a NULL on first call, init an hll_empty
     if (PG_ARGISNULL(0))
     {
+        int32 log2m = PG_GETARG_INT32(2);
+        int32 regwidth = PG_GETARG_INT32(3);
+        int64 expthresh = PG_GETARG_INT64(4);
+        int32 sparseon = PG_GETARG_INT32(5);
+
         msap = setup_multiset(aggctx);
+
+        check_modifiers(log2m, regwidth, expthresh, sparseon);
+
+        memset(msap, '\0', sizeof(multiset_t));
+
+        msap->ms_type = MST_EMPTY;
+        msap->ms_nbits = regwidth;
+        msap->ms_nregs = 1 << log2m;
+        msap->ms_log2nregs = log2m;
+        msap->ms_expthresh = expthresh;
+        msap->ms_sparseon = sparseon;
     }
     else
     {
@@ -2757,28 +2884,6 @@ hll_add_trans4(PG_FUNCTION_ARGS)
     {
         int64 val = PG_GETARG_INT64(1);
 
-        // Was the first argument uninitialized?
-        if (msap->ms_type == MST_UNINIT)
-        {
-            int32 log2m = PG_GETARG_INT32(2);
-            int32 regwidth = PG_GETARG_INT32(3);
-            int64 expthresh = PG_GETARG_INT64(4);
-            int32 sparseon = PG_GETARG_INT32(5);
-
-            multiset_t	ms;
-
-            check_modifiers(log2m, regwidth, expthresh, sparseon);
-
-            memset(msap, '\0', sizeof(ms));
-
-            msap->ms_type = MST_EMPTY;
-            msap->ms_nbits = regwidth;
-            msap->ms_nregs = 1 << log2m;
-            msap->ms_log2nregs = log2m;
-            msap->ms_expthresh = expthresh;
-            msap->ms_sparseon = sparseon;
-        }
-
         multiset_add(msap, val);
     }
 
@@ -2805,10 +2910,26 @@ hll_add_trans3(PG_FUNCTION_ARGS)
                 (errcode(ERRCODE_DATA_EXCEPTION),
                  errmsg("hll_add_trans3 outside transition context")));
 
-    // Is the first argument a NULL?
+    // If the first argument is a NULL on first call, init an hll_empty
     if (PG_ARGISNULL(0))
     {
+        int32 log2m = PG_GETARG_INT32(2);
+        int32 regwidth = PG_GETARG_INT32(3);
+        int64 expthresh = PG_GETARG_INT64(4);
+        int32 sparseon = g_default_sparseon;
+
         msap = setup_multiset(aggctx);
+
+        check_modifiers(log2m, regwidth, expthresh, sparseon);
+
+        memset(msap, '\0', sizeof(multiset_t));
+
+        msap->ms_type = MST_EMPTY;
+        msap->ms_nbits = regwidth;
+        msap->ms_nregs = 1 << log2m;
+        msap->ms_log2nregs = log2m;
+        msap->ms_expthresh = expthresh;
+        msap->ms_sparseon = sparseon;
     }
     else
     {
@@ -2820,28 +2941,6 @@ hll_add_trans3(PG_FUNCTION_ARGS)
     {
         int64 val = PG_GETARG_INT64(1);
 
-        // Was the first argument uninitialized?
-        if (msap->ms_type == MST_UNINIT)
-        {
-            int32 log2m = PG_GETARG_INT32(2);
-            int32 regwidth = PG_GETARG_INT32(3);
-            int64 expthresh = PG_GETARG_INT64(4);
-            int32 sparseon = g_default_sparseon;
-
-            multiset_t	ms;
-
-            check_modifiers(log2m, regwidth, expthresh, sparseon);
-
-            memset(msap, '\0', sizeof(ms));
-
-            msap->ms_type = MST_EMPTY;
-            msap->ms_nbits = regwidth;
-            msap->ms_nregs = 1 << log2m;
-            msap->ms_log2nregs = log2m;
-            msap->ms_expthresh = expthresh;
-            msap->ms_sparseon = sparseon;
-        }
-
         multiset_add(msap, val);
     }
 
@@ -2868,10 +2967,26 @@ hll_add_trans2(PG_FUNCTION_ARGS)
                 (errcode(ERRCODE_DATA_EXCEPTION),
                  errmsg("hll_add_trans2 outside transition context")));
 
-    // Is the first argument a NULL?
+    // If the first argument is a NULL on first call, init an hll_empty
     if (PG_ARGISNULL(0))
     {
+        int32 log2m = PG_GETARG_INT32(2);
+        int32 regwidth = PG_GETARG_INT32(3);
+        int64 expthresh = g_default_expthresh;
+        int32 sparseon = g_default_sparseon;
+
         msap = setup_multiset(aggctx);
+
+        check_modifiers(log2m, regwidth, expthresh, sparseon);
+
+        memset(msap, '\0', sizeof(multiset_t));
+
+        msap->ms_type = MST_EMPTY;
+        msap->ms_nbits = regwidth;
+        msap->ms_nregs = 1 << log2m;
+        msap->ms_log2nregs = log2m;
+        msap->ms_expthresh = expthresh;
+        msap->ms_sparseon = sparseon;
     }
     else
     {
@@ -2883,28 +2998,6 @@ hll_add_trans2(PG_FUNCTION_ARGS)
     {
         int64 val = PG_GETARG_INT64(1);
 
-        // Was the first argument uninitialized?
-        if (msap->ms_type == MST_UNINIT)
-        {
-            int32 log2m = PG_GETARG_INT32(2);
-            int32 regwidth = PG_GETARG_INT32(3);
-            int64 expthresh = g_default_expthresh;
-            int32 sparseon = g_default_sparseon;
-
-            multiset_t	ms;
-
-            check_modifiers(log2m, regwidth, expthresh, sparseon);
-
-            memset(msap, '\0', sizeof(ms));
-
-            msap->ms_type = MST_EMPTY;
-            msap->ms_nbits = regwidth;
-            msap->ms_nregs = 1 << log2m;
-            msap->ms_log2nregs = log2m;
-            msap->ms_expthresh = expthresh;
-            msap->ms_sparseon = sparseon;
-        }
-
         multiset_add(msap, val);
     }
 
@@ -2931,10 +3024,26 @@ hll_add_trans1(PG_FUNCTION_ARGS)
                 (errcode(ERRCODE_DATA_EXCEPTION),
                  errmsg("hll_add_trans1 outside transition context")));
 
-    // Is the first argument a NULL?
+    // If the first argument is a NULL on first call, init an hll_empty
     if (PG_ARGISNULL(0))
     {
+        int32 log2m = PG_GETARG_INT32(2);
+        int32 regwidth = g_default_regwidth;
+        int64 expthresh = g_default_expthresh;
+        int32 sparseon = g_default_sparseon;
+
         msap = setup_multiset(aggctx);
+
+        check_modifiers(log2m, regwidth, expthresh, sparseon);
+
+        memset(msap, '\0', sizeof(multiset_t));
+
+        msap->ms_type = MST_EMPTY;
+        msap->ms_nbits = regwidth;
+        msap->ms_nregs = 1 << log2m;
+        msap->ms_log2nregs = log2m;
+        msap->ms_expthresh = expthresh;
+        msap->ms_sparseon = sparseon;
     }
     else
     {
@@ -2946,28 +3055,6 @@ hll_add_trans1(PG_FUNCTION_ARGS)
     {
         int64 val = PG_GETARG_INT64(1);
 
-        // Was the first argument uninitialized?
-        if (msap->ms_type == MST_UNINIT)
-        {
-            int32 log2m = PG_GETARG_INT32(2);
-            int32 regwidth = g_default_regwidth;
-            int64 expthresh = g_default_expthresh;
-            int32 sparseon = g_default_sparseon;
-
-            multiset_t	ms;
-
-            check_modifiers(log2m, regwidth, expthresh, sparseon);
-
-            memset(msap, '\0', sizeof(ms));
-
-            msap->ms_type = MST_EMPTY;
-            msap->ms_nbits = regwidth;
-            msap->ms_nregs = 1 << log2m;
-            msap->ms_log2nregs = log2m;
-            msap->ms_expthresh = expthresh;
-            msap->ms_sparseon = sparseon;
-        }
-
         multiset_add(msap, val);
     }
 
@@ -2994,10 +3081,26 @@ hll_add_trans0(PG_FUNCTION_ARGS)
                 (errcode(ERRCODE_DATA_EXCEPTION),
                  errmsg("hll_add_trans0 outside transition context")));
 
-    // Is the first argument a NULL?
+    // If the first argument is a NULL on first call, init an hll_empty
     if (PG_ARGISNULL(0))
     {
+        int32 log2m = g_default_log2m;
+        int32 regwidth = g_default_regwidth;
+        int64 expthresh = g_default_expthresh;
+        int32 sparseon = g_default_sparseon;
+
         msap = setup_multiset(aggctx);
+
+        check_modifiers(log2m, regwidth, expthresh, sparseon);
+
+        memset(msap, '\0', sizeof(multiset_t));
+
+        msap->ms_type = MST_EMPTY;
+        msap->ms_nbits = regwidth;
+        msap->ms_nregs = 1 << log2m;
+        msap->ms_log2nregs = log2m;
+        msap->ms_expthresh = expthresh;
+        msap->ms_sparseon = sparseon;
     }
     else
     {
@@ -3009,28 +3112,6 @@ hll_add_trans0(PG_FUNCTION_ARGS)
     {
         int64 val = PG_GETARG_INT64(1);
 
-        // Was the first argument uninitialized?
-        if (msap->ms_type == MST_UNINIT)
-        {
-            int32 log2m = g_default_log2m;
-            int32 regwidth = g_default_regwidth;
-            int64 expthresh = g_default_expthresh;
-            int32 sparseon = g_default_sparseon;
-
-            multiset_t	ms;
-
-            check_modifiers(log2m, regwidth, expthresh, sparseon);
-
-            memset(msap, '\0', sizeof(ms));
-
-            msap->ms_type = MST_EMPTY;
-            msap->ms_nbits = regwidth;
-            msap->ms_nregs = 1 << log2m;
-            msap->ms_log2nregs = log2m;
-            msap->ms_expthresh = expthresh;
-            msap->ms_sparseon = sparseon;
-        }
-
         multiset_add(msap, val);
     }
 
@@ -3256,3 +3337,25 @@ hll_ceil_card_unpacked(PG_FUNCTION_ARGS)
         PG_RETURN_INT64(ceilval);
     }
 }
+
+PG_FUNCTION_INFO_V1(hll_recv);
+Datum hll_recv(PG_FUNCTION_ARGS);
+Datum
+hll_recv(PG_FUNCTION_ARGS)
+{
+    Datum dd = DirectFunctionCall1(bytearecv, PG_GETARG_DATUM(0));
+    return dd;
+}
+
+PG_FUNCTION_INFO_V1(hll_send);
+Datum hll_send(PG_FUNCTION_ARGS);
+Datum
+hll_send(PG_FUNCTION_ARGS)
+{
+    Datum dd = PG_GETARG_DATUM(0);
+    bytea* bp = DatumGetByteaP(dd);
+    StringInfoData buf;
+    pq_begintypsend(&buf);
+    pq_sendbytes(&buf, VARDATA(bp), VARSIZE(bp) - VARHDRSZ);
+    PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
